//--------------------------
// TAO - 2006 - Sylvain Huet
//--------------------------

#ifndef _BITMAP_
#define _BITMAP_

#include"param.h"

//#define ATTEN32

#ifdef ATTEN32
extern unsigned short attenh[256*32];
extern unsigned short attenm[256*64];
extern unsigned short attenl[256*32];

#define ATTENUATE(col,k) (attenh[((col>>3)&0x1f00)+(k)]+attenm[(((col)<<3)&0x3f00)+(k)]+attenl[(((col)<<8)&0x1f00)+(k)])
#else
extern short attenh[65536];
extern unsigned char attenl[65536];

#define ATTENUATE(col,k) (attenh[((col)&0xff00)+(k)]+attenl[(((col)<<8)&0xff00)+(k)])
#endif

#define ALPHA(dst,src,k) (ATTENUATE(src,k)+ATTENUATE(dst,255-(k)))


extern short attenh_[65536];
extern unsigned char attenl_[65536];


#define ATTENUATE_(col,k) (attenh_[((col)&0xff00)+(k)]+attenl_[(((col)<<8)&0xff00)+(k)])
#define ALPHA_(dst,src,k) (ATTENUATE_(src,k)+ATTENUATE_(dst,255-(k)))

#define COL_TITLE RGBTO16(255,87,45)

void attenuateInit();

int extstrlen(char* p);
void spaceremove(char* buf,char* inp);

class Filter
{
public:
	short hf[256];
	short lf[256];
	Filter(int col,int tx);
};


class Font
{
private:
	char* data;
	int nbchars;
	short* tabchar;

	int ascii[256];

	int* taboffs;

	char* bytestart;
	int bytenextline;
	int bytenblines;

	char* bitstart;
	int bitnextline;
	int bitnblines;


	int msbtoint(char *p);
	int msbtoword(char *p);

	int findchar(int c);

public:
	int liss;
	int height;
	Font();
	~Font();
	int load(char *filename);
	char* getchar_(int c,int* nextline,int* nblines,int *offsetbit,int *width);
	void setliss(int val);
	int strwidth(char *str,int len);
	int strheight(char *str,int len);
	int loaded();
};


class Bitmap
{
public:
	short* start;
	int width;
	int height;

	char* name;
	Bitmap *next;
	int count;
	int thread;

	Bitmap(int w, int h);
	~Bitmap();
	void drawplot(int x, int y, short col);
	void drawverticalline(int x, int y1, int y2, short col);
	void drawhorizontalline(int x1, int x2, int y, short col);
	void drawbox(int x, int y, int w, int h, short col);
	void drawboxf(int x, int y, int w, int h, short col);
	void drawline(int x1, int y1, int x2, int y2, short col);
	void drawlinebold(int x1, int y1, int x2, int y2, short col);

	void empty(short col);

	char* cuttext(char* src,int len,int* ndisplay,int* next,int* wdisplay,int* hdisplay,int wmax,Font *font,Font **nextfont);
	int textnblines(int w,char *src,int len,Font* font,int *htxt);

	void drawcharliss_(int x,int y,uchar *data,int nextline,int nblines,int width,short textcol,int backcol);
	void drawchar_(int x,int y,uchar *data,int offsetbit,int nextline,int nblines,int width,int textcol,int backcol);

	void drawstr_(int x,int y,char *src,int len,int* textcol,int* backcol,Font* font);
	int drawstrbox_(int x,int y,int w,int h,int type,char *src,int len,int textcol,int backcol,Font* font);
	int drawtextbox(int x,int y,int w,int h,int type,char* src,int len,int textcol,int backcol,Font* font);
	int drawtextboxEx(int x,int y,int w,int h,int type,char* src,int len,int textcol,int backcol,Font* font,Font* font2);

	int clip(int &xd,int &yd,Bitmap* src,int &xs,int &ys,int &w,int &h);
	void blit(int xd,int yd,Bitmap* src,int xs,int ys,int w,int h,int col);
	void blittransp(int xd,int yd,Bitmap* src,int xs,int ys,int w,int h,int col);
	int clipstretch(int &xd,int &yd,int &wd,int &hd,Bitmap* src,int &xs,int &ys,int &ws,int &hs);
	void stretch(int xd,int yd,int wd,int hd,Bitmap* src,int xs,int ys,int ws,int hs);

	void blitalphaconst(int xd,int yd,Bitmap* src,int xs,int ys,int w,int h,uchar alpha,int col);
	void blitshadow(int xd,int yd,Bitmap* src,int xs,int ys,int w,int h);

	int clipdoublesrc(int &xd,int &yd,Bitmap* src1,int &xs1,int &ys1,Bitmap* src2,int &xs2,int &ys2,int &w,int &h);
	void blitalpha(int xd,int yd,Bitmap* src,int xs,int ys,Bitmap* alpha,int xa,int ya,int w,int h);

};


#define ALIGN_HOR_MASK 3
#define ALIGN_LEFT 0
#define ALIGN_RIGHT 1
#define ALIGN_CENTER 2

#define ALIGN_VER_MASK 0x1c
#define ALIGN_TOP 0
#define ALIGN_BOTTOM 4
#define ALIGN_BASELINE 8
#define ALIGN_VCENTER 16

extern Font* font_normal;
extern Font* font_title;
extern Font* font_subtitle;
extern Font* font_note;
extern Font* font_key;
extern Font* font_underline;

#define FONT_TITLE 0
#define FONT_SUBTITLE 1
#define FONT_NORMAL 2
#define FONT_UNDERLINE 3
#define FONT_NOTE 4
#define FONT_KEY 5
#define FONT_SYSTEM 6


void initFont();
int fonth(Font *f);
#endif