//--------------------------
// TAO - 2006 - Sylvain Huet
//--------------------------

#include <windows.h>
#include <windowsx.h>
#include <wingdi.h>
#include <commctrl.h>
#include <commdlg.h>
#include <string.h>

#include "param.h"
#include "bitmap.h"
#include "file.h"

Font* font_normal;
Font* font_title;
Font* font_subtitle;
Font* font_note;
Font* font_key;
Font* font_underline;
Font* systemfont=NULL;

#define FONT_key "Bitstream Vera Sans_18_1.ft"
#define FONT_note "Tahoma_12_0.ft"
#define FONT_normal "Tahoma_12_1.ft"
#define FONT_subtitle "Tahoma_14_1.ft"
#define FONT_title "Bitstream Vera Sans_16_1.ft"
#define FONT_underline "Tahoma_12_5.ft"

#define FONT_MAX 7
Font* tabfont[FONT_MAX];

void initFont()
{
	font_title=new Font();
	font_title->load(FONT_title);
	font_title->setliss(1);

	font_subtitle=new Font();
	font_subtitle->load(FONT_subtitle);
	font_subtitle->setliss(1);

	font_normal=new Font();
	font_normal->load(FONT_normal);
	font_normal->setliss(1);

	font_underline=new Font();
	font_underline->load(FONT_underline);
	font_underline->setliss(1);

	font_note=new Font();
	font_note->load(FONT_note);
	font_note->setliss(0);

	font_key=new Font();
	font_key->load(FONT_key);
	font_key->setliss(1);

	systemfont=font_key;

	tabfont[FONT_TITLE]=font_title;
	tabfont[FONT_SUBTITLE]=font_subtitle;
	tabfont[FONT_NORMAL]=font_normal;
	tabfont[FONT_UNDERLINE]=font_underline;
	tabfont[FONT_NOTE]=font_note;
	tabfont[FONT_KEY]=font_key;
	tabfont[FONT_SYSTEM]=systemfont;
/*
	font_title=new Font();
	font_title->setliss(1);
	font_title->load("Tahoma_18_1.ft");

	font_subtitle=new Font();
	font_subtitle->setliss(1);
	font_subtitle->load("Tahoma_14_1.ft");

	font_normal=new Font();
	font_normal->setliss(1);
	font_normal->load("Tahoma_12_1.ft");

	font_note=new Font();
	font_note->setliss(1);
	font_note->load("Tahoma_12_0.ft");

	font_key=new Font();
	font_key->setliss(1);
	font_key->load("Arial_24_1.ft");

	systemfont=new Font();
	systemfont->setliss(0);
	systemfont->load("Arial_18_1.ft");
*/
}

int fonth(Font *font)
{
	if (!font) font=systemfont;
	if (!font) return 1;
	return font->height;
}

Font::Font()
{
	data=NULL;
	height=1;
}

Font::~Font()
{
	if (data) fileRelease(data);
}

int Font::msbtoint(char *p)
{
	return ((p[0]&255)<<24)+((p[1]&255)<<16)+((p[2]&255)<<8)+(p[3]&255);
}

int Font::msbtoword(char *p)
{
	return ((p[0]&255)<<8)+(p[1]&255);
}

int Font::findchar(int c)
{
	if (c<256) return ascii[c];

	int i;
	for(i=0;i<nbchars;i++)
		if (c==msbtoword((char*)&tabchar[i])) return i;
	return -1;
}

int Font::load(char *filename)
{
	data=fileLoad(filename);
	if (!data) return -1;

	int i=0;
	nbchars=msbtoint(data+i); i+=4;

	tabchar=(short*)(&data[i]);
	i+=nbchars*2;

	int j;
	for(j=0;j<256;j++) ascii[j]=-1;
	for(j=0;j<nbchars;j++)
	{
		int k=msbtoword((char*)&tabchar[j]);
		if (k<256) ascii[k]=j;
	}

	taboffs=(int*)(&data[i]);
	i+=(nbchars+1)*4;

	bytenextline=msbtoint(data+i); i+=4;
	bytenblines=msbtoint(data+i); i+=4;
	bytestart=(&data[i]);
	i+=bytenextline*bytenblines;

	bitnextline=msbtoint(data+i); i+=4;
	bitnblines=msbtoint(data+i); i+=4;
	bitstart=(&data[i]);
	i+=bitnextline*bitnblines;

	liss=1;
	height=bytenblines;
	return 0;
}

char* Font::getchar_(int c,int* nextline,int* nblines,int *offsetbit,int *width)
{
	int i=findchar(c);
	if (i<0) return NULL;
	int x0=msbtoint((char*)&taboffs[i]);
	int x1=msbtoint((char*)&taboffs[i+1]);
	*width=x1-x0;
	if (offsetbit)
	{
		*nextline=bitnextline;
		*nblines=bitnblines;
		*offsetbit=x0&7;
		return bitstart+(x0>>3);
	}
	*nextline=bytenextline;
	*nblines=bytenblines;
	return bytestart+x0;
}

void Font::setliss(int val)
{
	liss=val;
	height=(liss)?bytenblines:bitnblines;
}

int Font::loaded()
{
	return (data!=NULL);
}

int charhtoi(char c)
{
	if ((c>='0')&&(c<='9')) return c-'0';
	if ((c>='a')&&(c<='f')) return c+10-'a';
	if ((c>='A')&&(c<='F')) return c+10-'A';
	return 0;
}

int charsep(int c)
{

	if ((c>='0')&&(c<='9')) return 0;
	if ((c>='a')&&(c<='z')) return 0;
	if ((c>='A')&&(c<='Z')) return 0;
	if ((c>=128)&&(c<=255)) return 0;
	if (c=='(') return 0;
	return 1;
}

int srcgetchar(char *src,int* len)
{
	char c=*(src++);
	*len=1;
	if (c!='\\') return c&255;
	*len=2;
	c=*(src++);
	if (c=='n') return 10;
	if ((c=='$')||(c=='&')||(c=='*'))
	{
		*len=6;
		int val=(charhtoi(src[0])<<12)+(charhtoi(src[1])<<8)+(charhtoi(src[2])<<4)+charhtoi(src[3]);
		if (c=='$') val+=0x10000;
		else if (c=='&') val+=0x20000;
		else val+=0x30000;
		return val;
	}
	*len=5;
	return (charhtoi(c)<<12)+(charhtoi(src[0])<<8)+(charhtoi(src[1])<<4)+charhtoi(src[2]);
}

int Font::strwidth(char *str,int len)
{
	int w=0;
	int j;
	int lc;
	for(j=0;j<len;j++)
	{
		int c=srcgetchar(str+j,&lc);
		j+=lc-1;
		if (!(c&0xff0000))
		{
			int i=findchar(c);
			if (i>=0)
			{
				int x0=msbtoint((char*)&taboffs[i]);
				int x1=msbtoint((char*)&taboffs[i+1]);
				w+=x1-x0;
			}
		}
		else if ((c>>16)==3)
		{
			c&=0xffff;
			if (c<FONT_MAX) return w+tabfont[c]->strwidth(str+j+1,len-j-1);
		}
	}
	return w;
}

int Font::strheight(char *str,int len)
{
	if (liss) return bytenblines;
	return bitnblines;
}


void Bitmap::drawcharliss_(int x,int y,uchar *data,int nextline,int h,int w,short textcol,int backcol)
{
	int i,j;

	int offx=0;
	int offy=0;
	if (x<0) { w+=x; offx=-x; x=0; }
	if (x+w>width) w=width-x;
	if (y<0) { h+=y; offy=-y; y=0; }
	if (y+h>height) h=height-y;
	if ((w<=0)||(h<0)) return;

	short *buffer;
	buffer  = start + y*width + x;

	data+=offx+offy*nextline;

	if (backcol<0)
	for(j=0;j<h;j++)
	{
		uchar* d=data;
		short* p=buffer;
		for(i=0;i<w;i++)
		{
//			int zz=(d[i]*d[i])>>8;
			*p=ALPHA((*p),textcol,d[i]);
			p++;
		}
		buffer += width;
		data+=nextline;
	}
	else
	for(j=0;j<h;j++)
	{
		uchar* d=data;
		short *p=buffer;
		for(i=0;i<w;i++)
		{
			*p=ALPHA(backcol,textcol,d[i]);
			p++;
		}
		buffer += width;
		data+=nextline;
	}
}

void Bitmap::drawchar_(int x,int y,uchar *data,int offsetbit,int nextline,int h,int w,int textcol,int backcol)
{
	int i,j;

	int offx=0;
	int offy=0;
	if (x<0) { w+=x; offx=-x; x=0; }
	if (x+w>width) w=width-x;
	if (y<0) { h+=y; offy=-y; y=0; }
	if (y+h>height) h=height-y;
	if ((w<=0)||(h<0)) return;

	short *buffer;
	buffer  = start + y*width + x;

	offsetbit+=offx;

	data+=(offsetbit>>3)+offy*nextline;
	offsetbit&=7;
	offsetbit=128>>offsetbit;

	if (backcol<0)
	{
		if (textcol<0)
		for(j=0;j<h;j++)
		{
			uchar* d=data;
			int ofb=offsetbit;
			short *p=buffer;

			for(i=0;i<w;i++)
			{
				short g=(((((*p)&0x1f)+(((*p)>>5)&0x3f)+(((*p)>>11)&0x1f))+0x40)&0x7f)>>2;
				*p=((*d)&ofb)?((g<<11)+(g<<6)+g):(*p);
				p++;
				ofb>>=1;
				if (!ofb) { ofb=128; d++; }
			}
			buffer += width;
			data+=nextline;
		}
		else
		for(j=0;j<h;j++)
		{
			uchar* d=data;
			int ofb=offsetbit;
			short *p=buffer;

			for(i=0;i<w;i++)
			{
				*p=((*d)&ofb)?textcol:(*p);
				p++;
				ofb>>=1;
				if (!ofb) { ofb=128; d++; }
			}
			buffer += width;
			data+=nextline;
		}
	}
	else
	for(j=0;j<h;j++)
	{
		uchar* d=data;
		int ofb=offsetbit;
		short *p=buffer;

		for(i=0;i<w;i++)
		{
			*p=((*d)&ofb)?textcol:backcol;
			p++;
			ofb>>=1;
			if (!ofb) { ofb=128; d++; }
		}
		buffer += width;
		data+=nextline;
	}
}

void Bitmap::drawstr_(int x,int y,char *src,int len,int *textcol,int *backcol,Font* font)
{
	if (!font) font=systemfont;
	if (!font->loaded()) return;


	for(int i=0;i<len;i++)
	{
		int lc;
		int c=srcgetchar(src+i,&lc);
		i+=lc-1;

		int nextline,nblines,offsetbit,width;
		if (c&0xff0000)
		{
			int k=c>>16;
			c&=0xffff;
			if (k==1) *textcol=c;
			if (k==2) *backcol=c;
			if ((k==3)&&(c<FONT_MAX)) font=tabfont[c];
		}
		else
		{
			if (font->liss)
			{
				char* p=font->getchar_(c,&nextline,&nblines,NULL,&width);
				if (p)
				{
					drawcharliss_(x,y-font->height,(unsigned char*)p,nextline,nblines,width,*textcol,*backcol);
					x+=width;
				}
			}
			else
			{
				char* p=font->getchar_(c,&nextline,&nblines,&offsetbit,&width);
				if (p)
				{
					drawchar_(x,y-font->height,(unsigned char*)p,offsetbit,nextline,nblines,width,*textcol,*backcol);
					x+=width;
				}
			}
		}
	}
}



#include<stdio.h>

// dcoupe une ligne de texte : retourne le nombre de caractres
char* Bitmap::cuttext(char* src,int len,int* ndisplay,int* next,int* wdisplay,int* hdisplay,int wmax,Font *font,Font **nextfont)
{
	int offs=0;
	int offl=offs;
	int offcand=-1;
	int ws=0;
	int hs=0;
	int wcand;
	int hcand;
	int wc;
	int offsb;
	int nextline;
	int nblines;
	int* offsbit=NULL;
	if (!font->liss) offsbit=&offsb;
	font->getchar_(32,&nextline,&nblines,offsbit,&wc);

	*nextfont=font;

	// on avance jusqu'au premier caractre
	while((offs<len)&&(src[offs]==32)) offs++;
	if (offs>=len)
	{
		*next=len;
		*ndisplay=0;
		*wdisplay=0;
		*hdisplay=0;
		return src;
	}
	int first=offs;

	while(offs<len)
	{
		int lc;
		int c=srcgetchar(src+offs,&lc);
//		PRINTF("c=%d lc=%d\n",c,lc);
		if (c&0xff0000)
		{
			int k=c>>16;
			c&=0xffff;
			offs+=lc;
			if ((k==3)&&(c<FONT_MAX)) font=tabfont[c];
		}
		else
		{
			if (c==10)
			{
				*next=offs+1;
				*ndisplay=offs-first;
				*wdisplay=ws;
				*hdisplay=hs;
				*nextfont=font;
				return src+first;
			}

			font->getchar_(c,&nextline,&nblines,offsbit,&wc);
			if (ws+wc>wmax)
			{
				if (offcand>=0)
				{
					offs=offcand;
					ws=wcand;
					hs=hcand;
				}
				*next=offs;
				*ndisplay=offs-first;
				*wdisplay=ws;
				*hdisplay=hs;
				*nextfont=font;
				return src+first;
			}
			offs+=lc;
			ws+=wc;
			if (hs<font->height) hs=font->height;
			if (charsep(c))
			{
				offcand=offs;
				wcand=ws;
				hcand=hs;
				*nextfont=font;
			}
		}
	}
	*next=offs;
	*ndisplay=offs-first;
	*wdisplay=ws;
	*hdisplay=hs;
	*nextfont=font;
	return src+first;
}

int Bitmap::textnblines(int w,char *src,int len,Font* font,int *htxt)
{
	if (!font) font=systemfont;
	if (!font->loaded()) return 0;

	int offs=0;
	int nlines=0;
	int ht=0;

	while(offs<len)
	{
//		PRINTF("t:%c ",src[offs]);
		int ndisplay,next,wdisplay,hdisplay;
		char* pdisplay=cuttext(src+offs,len-offs,&ndisplay,&next,&wdisplay,&hdisplay,w,font,&font);
		ht+=hdisplay;
//		PRINTF(" ->%d\n",ht);
		offs+=next;
		nlines++;
	}
	if (htxt) *htxt=ht;
	return nlines;
}

int Bitmap::drawstrbox_(int x,int y,int w,int h,int type,char *src,int len,int textcol,int backcol,Font* font)
{
	if (!font) font=systemfont;
	if (!font->loaded()) return y;

	int offs=0;
	int offx=0;
	int ofy=0;

	if (type&ALIGN_VCENTER)
	{
		int dh;
		textnblines(w,src,len,font,&dh);
//		PRINTF("display %s\ndh=%d / h=%d\n",src,dh,h);
		if (dh<h) ofy+=(h-dh)>>1;
	}
	while((offs<len)&&(ofy<=h))
	{
		int ndisplay,next,wdisplay,hdisplay;
		Font* nextfont;
		char* pdisplay=cuttext(src+offs,len-offs,&ndisplay,&next,&wdisplay,&hdisplay,w,font,&nextfont);

		int offx=0;
		if ((type&ALIGN_HOR_MASK)==ALIGN_RIGHT) offx=w-wdisplay;
		if ((type&ALIGN_HOR_MASK)==ALIGN_CENTER) offx=(w-wdisplay)>>1;

		ofy+=hdisplay;
		if (y+ofy>=0) drawstr_(x+offx,y+ofy,pdisplay,ndisplay,&textcol,&backcol,font);
		font=nextfont;
		offs+=next;
	}
	return y+ofy;
}

int extstrlen(char* p)
{
	if (!p) return 0;
	int n=strlen(p);
	int x=0;
	int i=0;
	while((i<n)&&(p[i]))
	{
		int j;
		srcgetchar(p+i,&j);
		x++;
		i+=j;
	}
	return x;
}

void spaceremove(char* buf,char* inp)
{
	int i=0;
	int j=0;
	while(inp[i])
	{
		if((inp[i]!=32)&&(inp[i]!=34)&&(inp[i]!='')) buf[j++]=inp[i];
		i++;
	}
	buf[j++]=0;
}

