//--------------------------
// TAO - 2006 - Sylvain Huet
//--------------------------

#include <windows.h>
#include <stdio.h>
#include <string.h>

#include "param.h"
#include "control.h"
#include "bitmap.h"
#include "container.h"

Control::Control(Container *c,Control* f,int x,int y,int w,int h)
{
	container=c;
	father=f;
	if (!f) f=container->getmainctr();
	if (f)
	{
		brother=f->sons;
		f->sons=this;
	}
	else brother=NULL;
	sons=NULL;

	enable=1;
	xs=x; ys=y; ws=w; hs=h;
	oxs=oys=0;
	move_dx=move_dy=move_count=0;
	infinite=0;
}

Control::~Control()
{
//	PRINTF("delete control\n");
	container->deletectr(this);
	if (brother) delete brother;
	if (sons) delete sons;
}

void Control::setenable(int flag)
{
	enable=flag;
}

int Control::getenable()
{
	return enable;
}

void Control::localpos(int *x,int *y,int *w,int *h)
{
	if (x) *x=xs;
	if (y) *y=ys;
	if (w) *w=ws;
	if (h) *h=hs;
}
void Control::globalpos(int *x,int *y)
{
	int xx=0;
	int yy=0;
	Control* c=this;
	while(c)
	{
		xx+=c->xs;
		yy+=c->ys;
		c=c->father;
	}
	*x=xx;
	*y=yy;
}

void Control::setpos(int x,int y)
{
	xs=x;
	ys=y;
}

void Control::setsize(int w,int h)
{
	ws=w;
	hs=h;
}

void Control::getpos(int *x,int *y)
{
	*x=xs;
	*y=ys;
}

void Control::setoff(int ox,int oy)
{
	oxs=ox;
	oys=oy;
}


void Control::drawrec(Bitmap *bmp,int dx,int dy)
{
	if (brother) brother->drawrec(bmp,dx,dy);
	if (enable)
	{
		if (move_count)
		{
			xs+=move_dx;
			ys+=move_dy;
			move_count--;
		}

		int x=dx+xs;
		int y=dy+ys;

		draw(bmp,x-oxs,y-oys);

		if (sons) sons->drawrec(bmp,x,y);
	}
}

Control* Control::clicrec(int dx,int dy)
{
	Control* ctmp;

	if (enable)
	{
		int xc=dx-xs;
		int yc=dy-ys;
		if ((sons)&&(ctmp=sons->clicrec(xc,yc))) return ctmp;
		xc+=oxs;
		yc+=oys;
		if ((inside(xc,yc))&&(ctmp=clic(xc,yc))) return ctmp;
	}
	if (brother) return brother->clicrec(dx,dy);
	return NULL;
}

int Control::inside(int xc,int yc)
{
	if (infinite) return 1;
	if ((xc>=0)&&(xc<ws)&&(yc>=0)&&(yc<hs)) return 1;
	return 0;
}


void Control::unclicrec(int x,int y)
{
	int xg,yg;
	globalpos(&xg,&yg);
	unclic(x-xg+oxs,y-yg+oys);
}
void Control::moveclicrec(int x,int y)
{
	int xg,yg;
	globalpos(&xg,&yg);
	moveclic(x-xg+oxs,y-yg+oys);
}

Control* Control::clic(int x,int y)
{
	return NULL;
}
void Control::unclic(int x,int y)
{
}
void Control::moveclic(int x,int y)
{
//	PRINTF("move %d,%d\n",x,y);
}
void Control::draw(Bitmap *bmp,int dx,int dy)
{

	bmp->drawbox(dx,dy,5,5,(short)0xffff);
	bmp->drawbox(dx,dy,ws,hs,(short)0xffff);
}

void Control::goback()
{
	if (!father) return;
	Control* p=father->sons;
	Control* lp=NULL;
	while(p)
	{
		if (p==this)
		{
			if (lp) lp->brother=p->brother;
			else father->sons=p->brother;
			p=p->brother;
		}
		else
		{
			lp=p;
			p=p->brother;
		}
	}
	lp->brother=this;
	brother=NULL;
}

void Control::gofront()
{
	if (!father) return;
	Control* p=father->sons;
	Control* lp=NULL;
	while(p)
	{
		if (p==this)
		{
			if (lp) lp->brother=p->brother;
			else father->sons=p->brother;
			p->brother=NULL;
			p=NULL;
		}
		else
		{
			lp=p;
			p=p->brother;
		}
	}
	brother=father->sons;
	father->sons=this;
}

void Control::link(Control* newfather)
{
	if (father)
	{
		Control* p=father->sons;
		Control* lp=NULL;
		while(p)
		{
			if (p==this)
			{
				if (lp) lp->brother=p->brother;
				else father->sons=p->brother;
				p->brother=NULL;
				p=NULL;
			}
			else
			{
				lp=p;
				p=p->brother;
			}
		}
	}
	if (newfather)
	{
		father=newfather;
		brother=father->sons;
		father->sons=this;
	}
}

void Control::move(int dx,int dy,int count)
{
	move_dx=dx;
	move_dy=dy;
	move_count=count;
}

int Control::moving()
{
	return move_count;
}
