//--------------------------
// TAO - 2006 - Sylvain Huet
//--------------------------

#include <windows.h>
#include <stdio.h>
#include <string.h>

#include "param.h"
#include "control.h"
#include "bitmap.h"
#include "rsc.h"

Ctr_button::Ctr_button(Container *c,Control* f,int x,int y,int w,int h,int nstates):Control(c,f,x,y,w,h)
{
	int i;

	if (nstates<1) nstates=1;
	nb_state=nstates;
	mskstate=new int[nb_state];

	for(i=0;i<NB_LAYER;i++) layer[i].tp=LAYER_DISABLED;
	for(i=0;i<NB_SRCBMP;i++) srcbmp[i].bmp=NULL;
	for(i=0;i<nb_state;i++) mskstate[i]=0;


	cbfun=NULL;
	setlayer(0,0,0,0xfc00,0,LAYER_FLATBOX);
	setmask(0,0);
	state=0;

}

Ctr_button::~Ctr_button()
{
	int i;
	for(i=0;i<NB_SRCBMP;i++) if (srcbmp[i].bmp) rscUnuseit(srcbmp[i].bmp);
	delete mskstate;
}

Control* Ctr_button::clic(int x,int y)
{
//	PRINTF("button down\n");
	if (cbfun)
	{
		(*cbfun)(this,x,y,CBBUTTON_DOWN,userparam1,userparam2);
		return this;
	}
	return NULL;
}
void Ctr_button::unclic(int x,int y)
{
//	PRINTF("button up\n");
	if (cbfun)
	{
		if (inside(x,y)) (*cbfun)(this,x,y,CBBUTTON_UP,userparam1,userparam2);
		else (*cbfun)(this,x,y,CBBUTTON_UPOUTSIDE,userparam1,userparam2);
	}
}

void Ctr_button::draw(Bitmap *bmp,int dx,int dy)
{
	if (!nb_state) return;
	int x=dx;
	int y=dy;


	int msk=mskstate[state];
	int i=0;
	while(msk)
	{
		if (msk&1)
		{
			int t=layer[i].tp;
			if (t==LAYER_FLATBOX)	bmp->drawboxf(x,y,ws,hs,layer[i].coef1);
			if ((t==LAYER_BLIT)||(t==LAYER_BLITTRANSP)||(t==LAYER_BLITALPHA)||(t==LAYER_BLITSHADOW))
			{
				Srcbmp* s=&srcbmp[layer[i].src1];
				Bitmap* b=s->bmp;
				if (b)
				{
					if (s->animwidth==-1)
					{
						srcbmp[i].offh=((GetTickCount()>>7)&3)*hs;
					}
					if (t==LAYER_BLITTRANSP) bmp->blittransp(x,y,b,s->offw,srcbmp[i].offh,ws,hs,s->color);
					else if (t==LAYER_BLITALPHA) bmp->blitalpha(x,y,b,s->offw,srcbmp[i].offh,
						srcbmp[layer[i].src2].bmp,srcbmp[layer[i].src2].offw,srcbmp[i].offh,ws,hs);
					else if (t==LAYER_BLITSHADOW) bmp->blitshadow(x,y,b,s->offw,srcbmp[i].offh,ws,hs);
					else bmp->blit(x,y,b,s->offw,srcbmp[i].offh,ws,hs,s->color);
					if (s->animwidth>0)
					{
						s->offw+=s->animwidth;
						if (s->offw>=b->width) s->offw=0;
					}
				}
			}
		}
		i++;
		msk>>=1;
	}
}

Ctr_button* Ctr_button_create(Container *c,Control* f,int x,int y,int w,int h,int nstates)
{
	return new Ctr_button(c,f,x,y,w,h,nstates);
}

void Ctr_button::setsrcbmp(int i,Bitmap *bmp,int color,int anim)
{
	if ((i<0)||(i>=NB_SRCBMP)) return;
	srcbmp[i].bmp=bmp;
	srcbmp[i].color=color;
	srcbmp[i].animwidth=anim;
	srcbmp[i].offw=0;
	srcbmp[i].offh=0;
}

void Ctr_button::movesrcbmp(int i,int x,int y)
{
	if ((i<0)||(i>=NB_SRCBMP)) return;
	srcbmp[i].offw=x;
	srcbmp[i].offh=y;
}


void Ctr_button::setlayer(int i,int s1,int s2,int c1,int c2,int tp)
{
	if ((i<0)||(i>=NB_LAYER)) return;
	layer[i].src1=s1;
	layer[i].src2=s2;
	layer[i].coef1=c1;
	layer[i].coef2=c2;
	layer[i].tp=tp;
}

void Ctr_button::setmask(int i,int msk)
{
	if ((i<0)||(i>=nb_state)) return;
	mskstate[i]=msk;
}

void Ctr_button::setstate(int i)
{
	if ((i<0)||(i>=nb_state)) return;
	state=i;
}

void Ctr_button::setcb(void (*cb)(Ctr_button *b,int x,int y,int state,void* user1,void* user2),void* user1,void* user2)
{
	cbfun=cb;
	userparam1=user1;
	userparam2=user2;
}

