//--------------------------
// TAO - 2006 - Sylvain Huet
//--------------------------

#include <windows.h>
#include <stdio.h>
#include <string.h>

#include "param.h"
#include "control.h"
#include "bitmap.h"
#include "rsc.h"




#define LIST_SPEED 10

Listitem::Listitem()
{
	style=0;
	font=NULL;
	name=NULL;
	button=NULL;
	icon=NULL;
	down=0;
	tx=0;
	tstyle=0;
	tcol=COL_TITLE;
	cbfun=NULL;
	score=0;
}

Listitem::~Listitem()
{
	if (name) delete name;
	rscUnuseit(button);
	rscUnuseit(icon);
}

void Listitem::setname(char* n)
{
	if (name) delete name;
	name=new char[strlen(n)+1];
	strcpy(name,n);
}
char* Listitem::getname()
{
	return name;
}

void Listitem::settext(int x,int style,int col)
{
	tx=x;
	tstyle=style;
	tcol=col;
}

void Listitem::setbutton(char* filename)
{
	if (button) rscUnuseit(button);
	button=rscGet(filename);
}
void Listitem::seticon(char* filename)
{
	if (icon) rscUnuseit(icon);
	icon=rscGet(filename);
}


void Listitem::seticonpos(int x,int y,int xs,int ys,int w,int h)
{
	ix=x; iy=y;
	ixs=xs; iys=ys;
	iw=w; ih=h;
}

void Listitem::setdown(int d)
{
	down=d;
}

int Listitem::getdown()
{
	return down;
}

void Listitem::setstyle(int s)
{
	style=s;
}

void Listitem::setscore(int s)
{
	score=s;
}

void Listitem::setfont(Font* f)
{
	font=f;
}

void Listitem::setcb(CBLIST cb,void* user1,void* user2)
{
	cbfun=cb;
	userparam1=user1;
	userparam2=user2;
}

void Listitem::callcb()
{
	if (cbfun) (*cbfun)(this,userparam1,userparam2);
}

void Listitem::draw(Bitmap* bmp,int x,int y,int w,int h)
{
	if ((style&LITEM_BUTTON)&&(button)) bmp->blit(x,y,button,0,(down)?h:0,w,h,0x1f);
	if ((style&LITEM_ICON)&&(icon)) bmp->blit(x+ix,y+iy,icon,ixs,iys,iw,ih,0x1f);
	if ((style&LITEM_TEXT)&&(name))
			bmp->drawtextbox(x+tx,y,w-tx,h,tstyle,name,strlen(name),(down&&cbfun)?0xffff:tcol,-1,font);
/*	{	char buf[128];
		sprintf(buf,"%d-%s",score,name);
		bmp->drawtextbox(x+tx,y,w-tx,h,tstyle,buf,strlen(buf),(down&&cbfun)?0xffff:tcol,-1,font);
	}
*/
}


Ctr_list::Ctr_list(Container *c,Control* f,int x,int y,int w,int h,int dy):Control(c,f,x,y,w,h)
{
	list=new Bitmap(w,h);
	items=NULL;
	dh=dy;
	anim=0;
	empty();
}

Ctr_list::~Ctr_list()
{
	empty();
	if (list) delete list;
}

Control* Ctr_list::clic(int x,int y)
{
//	PRINTF("button down\n");
	y=first+y/dh;
	if ((y>=0)&&(y<nitems))
	{
		items[y]->setdown(1);
		refresh();
		return this;
	}
	return NULL;
}

void Ctr_list::unclic(int x,int y)
{
//	PRINTF("button up\n");
	int cb=0;
	y=first+y/dh;
	if ((y>=0)&&(y<nitems))
	{
		if (items[y]->getdown()) cb=1;
	}
	for(int i=0;i<nitems;i++) items[i]->setdown(0);
	refresh();
	if (cb)	items[y]->callcb();
}

Bitmap* Ctr_list::getlist()
{
	return list;
}

void Ctr_list::refresh()
{

	list->empty(0x1f);

	int i,y;

	i=first;
	y=anim;
	while(y>0)
	{
		i--;
		y-=dh;
	}
	while(y+dh<0)
	{
		i++;
		y+=dh;
	}
//	PRINTF("y=%d i=%d nitems=%d hs=%d\n",y,i,nitems,hs);
	while((y<hs)&&(i<nitems))
	{
//		PRINTF("drawitem %d\n",i);
		if (i>=0) items[i]->draw(list,0,y,ws,dh);
		i++;
		y+=dh;
	}
}

void Ctr_list::empty()
{
	if (items)
	{
		for(int i=0;i<maxitems;i++) delete items[i];
		delete items;
	}
	nitems=0;
	maxitems=0;
	items=NULL;
	first=0;
	anim=0;
}

int Ctr_list::getnitems()
{
	return nitems;
}

Listitem* Ctr_list::getnthitem(int i)
{
	if ((i<0)||(i>=nitems)) return NULL;
	return items[i];
}

typedef Listitem* listitem;

Listitem* Ctr_list::getnextitem()
{
	if (nitems>=maxitems)
	{
		int mx=maxitems+10;
		Listitem** nl=new listitem[mx];
		for(int i=0;i<mx;i++)
			if (i<maxitems) nl[i]=items[i];
			else nl[i]=new Listitem;
		delete items;
		items=nl;
		maxitems=mx;
	}
	return items[nitems++];
}

void Ctr_list::sortname(int slimit)
{
	int cont=1;
	while(cont)
	{
		cont=0;
		for(int i=0;i<nitems-1;i++)
		{
			int exg=0;
			if ((items[i]->score<slimit)&&(items[i+1]->score>=slimit)) exg=1;
			else if ((items[i]->score<slimit)&&(items[i+1]->score<slimit)&&(strcmpi(items[i]->getname(),items[i+1]->getname())>0) ) exg=1;
			else if ((items[i]->score>=slimit)&&(items[i+1]->score>=slimit)&&(strcmpi(items[i]->getname(),items[i+1]->getname())>0) ) exg=1;
			if (exg)
			{
//				PRINTF("exg %d,%d %d-%s / %d-%s\n",i,i+1,items[i]->score,items[i]->getname(),items[i+1]->score,items[i+1]->getname());
				Listitem* li=items[i];
				items[i]=items[i+1];
				items[i+1]=li;
				cont=1;
			}
		}
	}
}

void Ctr_list::scroll(int delta)
{
	scrollto(first+delta);
}
void Ctr_list::scrollto(int i)
{
	if (i<0) i=0;
	if (i+(hs/dh)>nitems) i=nitems-(hs/dh);
	if (i<0) i=0;
	anim=(i-first)*dh;
	first=i;
	refresh();
}
void Ctr_list::setscroll(int i)
{
	if (i<0) i=0;
	if (i+(hs/dh)>nitems) i=nitems-(hs/dh);
	if (i<0) i=0;
	anim=0;
	first=i;
	refresh();
}

int Ctr_list::mayscroll(int i)
{
	i+=first;
	if (i<0) return 0;
	if (i+(hs/dh)>nitems) return 0;
	return 1;
}

int Ctr_list::getscroll()
{
	return first;
}

void Ctr_list::draw(Bitmap *bmp,int dx,int dy)
{
	if (anim)
	{
		if (anim<0)
		{
			anim+=LIST_SPEED;
			if (anim>0) anim=0;
		}
		else if (anim>0)
		{
			anim-=LIST_SPEED;
			if (anim<0) anim=0;
		}
		refresh();
	}
	bmp->blit(dx,dy,list,0,0,ws,hs,0x1f);
}

Ctr_list* Ctr_list_create(Container *c,Control* f,int x,int y,int w,int h,int dy)
{
	return new Ctr_list(c,f,x,y,w,h,dy);
}
