//--------------------------
// TAO - 2006 - Sylvain Huet
//--------------------------

#include <windows.h>
#include <stdio.h>
#include <string.h>

#include "param.h"
#include "file.h"
#include "control.h"
#include "interf.h"
#include "bitmap.h"
#include "rsc.h"
#include "poi.h"
#include "metro.h"

#include "raw/raw.h"

#define SCALEMIN 1
#define SCALEMAX 6
#define SCALEHIGH 6

#define ACCEPTABLE_MOVE_ON_CLIC 30

#define SUBLOADSIZE ((400*400)>>2)

#define MAPCOL RGBTO16(248,232,208)

#define AREAS_NB 3
const int area_x[AREAS_NB]={-1330,-8551,24820};
const int area_y[AREAS_NB]={-553,5584,1588};
const int area_w[AREAS_NB]={12000,3067,2173};
const int area_h[AREAS_NB]={6932,2408,2183};
#define AREAS_LIMIT 2000


#define ZONE_NB 22
const int zones[ZONE_NB*4]=
{
8629,5718,1788,532,-131,4198,1620,1312,6985,3962,3592,1888,-1311,2294,592,1480,-819,1838,564,2200,
-323,1558,1784,2756,7149,1330,884,4436,409,1242,1196,600,-103,950,980,524,1333,890,952,4704,
-435,690,636,496,-859,478,512,652,2145,138,1216,5888,6365,-54,940,5876,3325,-158,3280,6516,
-6863,6268,1360,1724,-8443,5956,2156,1580,-8019,5600,1540,1988,25104,2996,1108,748,
25372,2360,792,1056,25076,1852,1412,928,25460,1676,996,1172
};

int Ctr_map::isinarea(int x,int y)
{
	for(int i=0;i<ZONE_NB*4;i+=4)
		if ((x>=zones[i])&&(x<zones[i]+zones[i+2])&&(y>=zones[i+1])&&(y<zones[i+1]+zones[i+3])) return i;
	return -1;
}

void Ctr_map::middlearea(int i,int* x,int* y)
{
	*x=*y=0;
	if ((i<0)||(i>=AREAS_NB)) return;
	*x=area_x[i]+area_w[i]/2;
	*y=area_y[i]+area_h[i]/2;
}

int Ctr_map::isinarea(int x,int y,int limit)
{
	int ia=-1;
	int id=0;

	for(int i=0;i<AREAS_NB;i++)
	{
		int d=0;
		if ((x>=area_x[i])&&(x<area_x[i]+area_w[i])&&(y>=area_y[i])&&(y<area_y[i]+area_h[i])) return i;
		if (x<area_x[i]) d+=area_x[i]-x;
		else if (x>=area_x[i]+area_w[i]) d+=x-(area_x[i]+area_w[i]);
		if (y<area_y[i]) d+=area_y[i]-y;
		else if (y>=area_y[i]+area_w[i]) d+=y-(area_y[i]+area_h[i]);
		if ((d<limit)&&((ia<0)||(d<id)))
		{
			ia=i;
			id=d;
		}
	}
	return ia;
}

Ctr_map::Ctr_map(Container *c,Control* f,int x,int y,int w,int h,int wm,int hm):Control(c,f,x,y,w,h)
{
	int i,j;


	infinite=1;
	shell=new Control(container,(Control*)this,0,0,0,0);
	Bitmap* gif=rscGet("icons/pictos_poi.r");
	Bitmap* gif1000=rscGet("icons/pictos_monuments.r");

	for(i=0;i<4;i++)
	{
		submap[i]=new Submap;

		submap[i]->src=new Bitmap(wm,hm);
		submap[i]->xm=-1000;
		submap[i]->ym=-1000;
		for(j=0;j<MAP_MAXPOI;j++) submap[i]->poitab[j]=NULL;
		for(j=MAP_MAXPOI-1;j>=0;j--)
		{
			Ctr_button* poi=Ctr_button_create(container,shell,0,0,24,24,2);	// attention  la taille
			poi->setoff(12,12);
			poi->setsrcbmp(0,gif,0x1f,-1);
			poi->setlayer(0,0,0,0,0,LAYER_BLIT);
			poi->setmask(0,1);
			poi->setsrcbmp(1,gif1000,RGBTO16(220,220,220),-1);
			poi->setlayer(1,1,0,0,0,LAYER_BLIT);
			poi->setmask(1,2);
			poi->setstate(0);
			poi->setenable(0);
			submap[i]->poibtab[j]=poi;
		}
		submap[i]->file=0;
		submap[i]->offload=0;
	}

	scale=SCALEMIN+1;

	zoomdir=1;
	zoomk=0;

	dwm=wm;
	dhm=hm;

	xm=-1000;
	ym=-1000;

	lines=Ctr_lines_create(container,(Control*)this,0,0,0,0);
	lines->setscale(scale-SCALEMIN);
	

	gpstrack=0;
	target=0;

	gps=Ctr_button_create(container,(Control*)this,0,0,52,50,3);

	gps->setsrcbmp(0,rscGet("interf/gps/cible_gps1.r"),0x1f,0);
	gps->setlayer(0,0,0,0,0,LAYER_BLIT);
	gps->setsrcbmp(1,rscGet("interf/gps/cible_gps2.r"),0x1f,0);
	gps->setlayer(1,1,0,0,0,LAYER_BLIT);
	gps->setsrcbmp(2,rscGet("interf/gps/cible_gps3.r"),0x1f,0);
	gps->setlayer(2,2,0,0,0,LAYER_BLIT);

	
	gps->setoff(26,25);
	gps->setmask(0,2);
	gpsnotify(0,x,y);
}

int Ctr_map::xcoord(int x)
{
	if (scale!=SCALEHIGH) x=(x>>(scale-SCALEMIN));
	else x=106+(x-6965)*(68-106)/(607-6975);
	return x;
}
int Ctr_map::ycoord(int y)
{
	if (scale!=SCALEHIGH) y=(y>>(scale-SCALEMIN));
	else y=143+(y-452)*(171-143)/(4947-452);
	return y;
}

int Ctr_map::absxcoord(int x)
{
	if (scale!=SCALEHIGH) x=(x<<(scale-SCALEMIN));
	else x=6965+(x-106)*(607-6975)/(68-106);
	return x;
}
int Ctr_map::absycoord(int y)
{
	if (scale!=SCALEHIGH) y=(y<<(scale-SCALEMIN));
	else y=452+(y-143)*(4947-452)/(171-143);
	return y;
}

void Ctr_map::clicpoi(Poi* k)
{
//	PRINTF("clic poi %s\n",poilist->getname(k));
	
	ipoi->poigo(k,-1,NULL,NULL);
}

void icbpoi(Ctr_button *b,int x,int y,int state,void* user,void* user2)
{
	Poi* kpoi=(Poi*)user2;
	Ctr_map* m=(Ctr_map*)user;
	if ((state==CBBUTTON_UP)&&(kpoi)) m->clicpoi(kpoi);
}

void Ctr_map::updatepoi(int x,int y)
{
	Poi* tb[MAP_MAXPOI];
	int i,j,n;

//	PRINTF("#########UPDATEPOI");

	int nc=x+y+y;

	Poi** poitab_=submap[nc]->poitab;
	Ctr_button** poibtab_=submap[nc]->poibtab;

	n=poilist->getbest(tb,MAP_MAXPOI,absxcoord((xm+x)*dwm),absxcoord((xm+x+1)*dwm),
		absycoord((ym+y)*dhm),absycoord((ym+y+1)*dhm),scale,(scale-SCALEMIN));
//	PRINTF("n=%d\n",n);
	for(i=0;i<n;i++) tb[i]->tmp=1;
	for(i=0;i<MAP_MAXPOI;i++)
		if (poitab_[i])
		{
			if (!poitab_[i]->tmp)
			{
//				PRINTF("delete poi %x\n",poibtab_[i]);
//				PRINTF(": %d,%d (shell %d,%d)\n",poibtab_[i]->xs,poibtab_[i]->ys,this->xs,this->ys);
				poibtab_[i]->setenable(0);
				poibtab_[i]->setcb(icbpoi,(void*)this,(void*)NULL);
				poitab_[i]=NULL;
			}
			else poitab_[i]->tmp++;
		}
		else
		{
//			PRINTF("disable poi\n");
			poibtab_[i]->setenable(0);
		}

	j=0;
	int itb=0;
	for(i=0;i<n;i++)
		if (tb[i]->tmp==1)
		{
			PRINTF("create poi %x %s\n",tb[i],poilist->getname(tb[i]));

			while(poitab_[j]) j++;
			poitab_[j]=tb[i];

			while((itb<MAP_MAXPOI)&&(poibtab_[itb]->getenable())) itb++;
			if (itb<MAP_MAXPOI)
			{
				int ic=tb[i]->icon;

				Ctr_button* poi=poibtab_[itb];
				poi->xabs=tb[i]->x;
				poi->yabs=tb[i]->y;
				poi->setpos(xcoord(poilist->getpoix(tb[i])),ycoord(poilist->getpoiy(tb[i])));
				poi->setenable(1);
				if (ic<1000)
				{
					poi->movesrcbmp(0,ic*24,0);
					poi->setsize(24,24);
					poi->setoff(12,12);
					poi->setstate(0);
				}
				else
				{
					poi->movesrcbmp(1,(ic-1000)*36,0);
					poi->setsize(36,36);
					poi->setoff(18,18);
					poi->setstate(1);
				}
//				PRINTF("done\n");

				poi->setcb(icbpoi,(void*)this,(void*)poitab_[j]);
			}
//			PRINTF("create poi %x\n",poi);
		}
	for(i=0;i<n;i++) tb[i]->tmp=0;
}

void Ctr_map::resetpoi()
{
//	PRINTF("#########RESETPOI");
	for(int y=0;y<2;y++)
		for(int x=0;x<2;x++)
			for(int i=0;i<MAP_MAXPOI;i++) submap[x+y+y]->poitab[i]=NULL;
	updatepoi();
}


void Ctr_map::updatepoi()
{
	for(int y=0;y<2;y++)
		for(int x=0;x<2;x++)
			updatepoi(x,y);
}

void Ctr_map::gpsnotify(int state,int x,int y)
{
	gpsstate=state;
	gpsx=x;
	gpsy=y;

	if ((isinarea(gpsx,gpsy)<0)&&(gpstrack))
	{
		gpstrack=target=0;
	}
	gps->xabs=x;
	gps->yabs=y;
	x=xcoord(x);
	y=ycoord(y);
	gps->setpos(x,y);
}

void Ctr_map::gpssettrack(int tracker)
{
//	if (!gpsstate) tracker=0;
	if (isinarea(gpsx,gpsy)<0) return;

	gpstrack=tracker;
	target=gpstrack;
}

void Ctr_map::setscale(int newscale)
{
	if (newscale==scale) return;

	PRINTF("newscale=%d\n",newscale);
	PRINTF("xs=%d, ys=%d\n",xs,ys);
	xs-=(WIDTH>>1);
	ys-=(HEIGHT>>1);

	xs=absxcoord(xs);
	ys=absycoord(ys);
	PRINTF("xs=%d, ys=%d\n",xs,ys);
	if (newscale==SCALEHIGH)
	{
		lastxs=xs;
		lastys=ys;
	}
	if (scale==SCALEHIGH)
	{
		xs=lastxs;
		ys=lastys;
	}
	scale=newscale;
	xs=xcoord(xs);
	ys=ycoord(ys);
	xs+=(WIDTH>>1);
	ys+=(HEIGHT>>1);
	if (newscale==SCALEHIGH)
	{
		xs=0;
		ys=0;
	}
	PRINTF("xs=%d, ys=%d\n",xs,ys);
	Control* p=shell->sons;
	while(p)
	{
		p->xs=xcoord(p->xabs);
		p->ys=ycoord(p->yabs);
		if (p==gps) PRINTF("move gps %dx%d -> %dx%d\n",p->xabs,p->yabs,p->xs,p->ys);
		p=p->brother;
	}
	gps->xs=xcoord(gps->xabs);
	gps->ys=ycoord(gps->yabs);


	if (scale==SCALEMIN)
	{
		gps->setsize(78,75);
		gps->setoff(39,37);
		gps->setmask(0,4);
	}
	else if (scale==SCALEMIN+1)
	{
		gps->setsize(52,50);
		gps->setoff(26,25);
		gps->setmask(0,2);
	}
	else
	{
		gps->setsize(26,25);
		gps->setoff(13,13);
		gps->setmask(0,1);
	}
	lines->setscale(scale-SCALEMIN);

	for(int i=0;i<4;i++)
	{
		submap[i]->xm=-1000;
		submap[i]->ym=-1000;
	}

	xm=-1000;
	ym=-1000;
}

void Ctr_map::moveto(int x,int y,int delay)
{
	if (isinarea(x,y)>=0)
	{
		gpstrack=target=0;
		target=delay;
		targetx=x;
		targety=y;
	}
}

Ctr_map::~Ctr_map()
{
	int i;
	for(i=0;i<4;i++)
	{
		delete submap[i]->src;
		if (submap[i]->file) stdcloseRaw(submap[i]->file);
		delete submap[i];
	}
}

void Ctr_map::reload(int xxm,int yym)
{
	if ((xxm==xm+1)||(xxm==xm-1))
	{
		Submap* s;
		s=submap[0];submap[0]=submap[1];submap[1]=s;
		s=submap[2];submap[2]=submap[3];submap[3]=s;
	}
	if ((yym==ym+1)||(yym==ym-1))
	{
		Submap* s;
		s=submap[0];submap[0]=submap[2];submap[2]=s;
		s=submap[1];submap[1]=submap[3];submap[3]=s;
	}

	xm=xxm;
	ym=yym;
	for(int j=0;j<2;j++)
		for(int i=0;i<2;i++)
		{
			int n=i+j+j;
			if ((submap[n]->xm!=xxm+i)||(submap[n]->ym!=yym+j))
			{
				if (submap[n]->file) stdcloseRaw(submap[n]->file);
				submap[n]->file=0;
				submap[n]->src->empty((short)MAPCOL);
				submap[n]->xm=xxm+i;
				submap[n]->ym=yym+j;
				char buf[MAX_FILENAME];
#ifdef HASHFILE
				char buf2[64];
				sprintf(buf2,"paris_%02d_%dx%d.r",scale,500+xxm+i,500+yym+j);
				sprintf(buf,"map_/%02d/%s",fileHash(buf2,HASHFILE),buf2);
#else
				sprintf(buf,"map/paris_%02d_%dx%d.r",scale,500+xxm+i,500+yym+j);
#endif
				PRINTF("###############load %s\n",buf);
				submap[n]->file=stdopenRaw(buf);
				submap[n]->offload=0;
				updatepoi(i,j);
			}
		}
}

void Ctr_map::tryload()
{
	int i;
	for(i=0;i<4;i++) if (submap[i]->file)
	{
		submap[i]->offload=stdreadRaw(submap[i]->file,submap[i]->src,submap[i]->offload,SUBLOADSIZE);
		if (submap[i]->offload>=dwm*dhm*2)
		{
			stdcloseRaw(submap[i]->file);
			submap[i]->file=0;
		}
	}
}

Control* Ctr_map::clic(int x,int y)
{
	PRINTF("button down\n");
	if (scale==SCALEHIGH)
	{
		int xa=absxcoord(x-xs);
		int ya=absycoord(y-ys);
		int a=isinarea(xa,ya,AREAS_LIMIT);
		if (a>=0)
		{
			/*if (isinarea(xa,ya)<0)*/ middlearea(a,&xa,&ya);
			setscale(scale-3);
			xs=(WIDTH>>1)-(xcoord(xa));
			ys=(HEIGHT>>1)-(ycoord(ya));
			xclic=-1;
			zoomrun=0;
			gpstrack=target=0;
			shell->setenable(1);
			lines->setenable(1);
			gps->setenable(1);
		}
		return this;
	}
	if ((zoomk)||(scale==SCALEHIGH))
	{
		xclic=-1;
		return this;
	}
	xclic=x; yclic=y;
	move=0;
	return this;
}
void Ctr_map::unclic(int x,int y)
{
	PRINTF("button up\n");
	PRINTF("moved %d\n",move);
	if (move<ACCEPTABLE_MOVE_ON_CLIC)
	{
		int station=metro->hotspot(absxcoord(xclic),absycoord(yclic),scale-SCALEMIN,10<<(scale-SCALEMIN));
		if (station>=0)
		{
			interfActive(ITF_STATION,1);
			istation->redrawbar(station);
		}
	}
	xclic=-1;
}

void Ctr_map::moveclic(int x,int y)
{
	if (xclic==-1) return;
	if ((target)&&(!gpstrack)) return;

	int dx=x-xclic;
	int dy=y-yclic;
	move+=((dx<0)?-dx:dx)+((dy<0)?-dy:dy);

	if (isinarea(absxcoord((WIDTH>>1)-dx-xs),absycoord((HEIGHT>>1)-dy-ys))>=0)
	{
		xs+=dx;
		ys+=dy;
		gpssettrack(0);
	}
	else if (isinarea(absxcoord((WIDTH>>1)-dx-xs),absycoord((HEIGHT>>1)-ys))>=0)
	{
		xs+=dx;
		gpssettrack(0);
	}
	else if (isinarea(absxcoord((WIDTH>>1)-xs),absycoord((HEIGHT>>1)-dy-ys))>=0)
	{
		ys+=dy;
		gpssettrack(0);
	}
//	else PRINTF("not in area\n");


}

int mye(int x,int y)
{
	int i=0;
	while(1)
	{
		if (x<0) { x+=y; i--; }
		else if (x>=y) {x-=y; i++; }
		else return i;
	}
}

void Ctr_map::draw(Bitmap *bmp,int dx,int dy)
{
	int x=-(dx);
	int y=-(dy);

	int xt=x-xm*dwm;
	int yt=y-ym*dhm;

	if ((xt<0)||(xt>=dwm)||(yt<0)||(yt>=dhm))
	{
		reload(mye(x,dwm),mye(y,dhm));
		xt=x-xm*dwm;
		yt=y-ym*dhm;
	}
	else tryload();
	if (zoomk)
	{
		int xt2=(((256+zoomk)*(xt+(WIDTH>>1)))>>8)-(WIDTH>>1);	// on cherche la coordonne du coin gauche
		int yt2=(((256+zoomk)*(yt+(HEIGHT>>1)))>>8)-(HEIGHT>>1);

		int dswm=dwm*(256+zoomk)>>8; // on calcule la largeur stretche
		int dshm=dhm*(256+zoomk)>>8;


		for(int j=0;j<2;j++)
			for(int i=0;i<2;i++)
			{
				Bitmap *p=submap[i+(j<<1)]->src;
				if (p) bmp->stretch(-xt2+i*dswm,-yt2+j*dshm,dswm,dshm,p,0,0,dwm,dhm);
				else bmp->drawboxf(-xt2+i*dswm,-yt2+j*dshm,dswm,dshm,(short)MAPCOL);
			}
		zoomk-=zoomdir*16;
		if (zoomk>=256)
		{
			zoomk=0;
			setscale(scale-1);
			shell->setenable(1);
			lines->setenable(1);
			gps->setenable(1);
			if (zoomrun) zoom(0);
		}
		else if (zoomk<0)
		{
			zoomk=0;
			shell->setenable(1);
			lines->setenable(1);
			gps->setenable(1);
			if (zoomrun) zoom(0);
		}

	}
	else
	{
		for(int j=0;j<2;j++)
			for(int i=0;i<2;i++)
			{
				Bitmap *p=submap[i+(j<<1)]->src;
				if (p) bmp->blit(-xt+i*dwm,-yt+j*dhm,p,0,0,dwm,dhm,-1);
				else bmp->drawboxf(-xt+i*dwm,-yt+j*dhm,dwm,dhm,(short)MAPCOL);
			}
		if (scale!=SCALEHIGH)
		{
			if ((gpstrack)&&(gpsx)&&(gpsy))
			{
				targetx=gpsx;
				targety=gpsy;
			}
			if (target>1) target--;
			else if (target)
			{
				if (isinarea(targetx,targety)<0) target=gpstrack=0;

				int gx=(WIDTH>>1)-(xcoord(targetx));
				int gy=(HEIGHT>>1)-(ycoord(targety));
				if ((gx!=xs)||(gy!=ys))
				{
					if ((gx<xs+16)&&(gx>xs-16)) xs+=(xs==gx)?0:(xs<gx)?1:-1;
					else xs=(gx+xs*15)>>4;
					if ((gy<ys+16)&&(gy>ys-16)) ys+=(ys==gy)?0:(ys<gy)?1:-1;
					else ys=(gy+ys*15)>>4;
				}
				else if (!gpstrack) target=0;
//				PRINTF("%d,%d -> %d,%d\n",xs,ys,gx,gy);
			}
		}
/*		else
		{
			for(int i=0;i<AREAS_NB;i++)
			{
				int ax=xcoord(area_x[i]);
				int ay=ycoord(area_y[i]);
				int ax1=xcoord(area_x[i]+area_w[i]);
				int ay1=ycoord(area_y[i]+area_h[i]);
	//			PRINTF("area %d,%d - %d,%d -%d,%d\n",ax,ay,ax1,ay1,-xt+i*dwm,-yt+j*dhm);
//				bmp->drawbox(-xt+ax,-yt+ay,ax1-ax,ay1-ay,0xf000);
				ax=xcoord(area_x[i]-AREAS_LIMIT);
				ay=ycoord(area_y[i]-AREAS_LIMIT);
				ax1=xcoord(area_x[i]+area_w[i]+AREAS_LIMIT);
				ay1=ycoord(area_y[i]+area_h[i]+AREAS_LIMIT);
				bmp->drawbox(-xt+ax,-yt+ay,ax1-ax,ay1-ay,0x1f);

			}
		}
*/
/*		if ((xclic!=-1)&&(countdown)&&(GetTickCount()-countdown>1000))
		{
			countdown=0;
			int station=metro->hotspot(absxcoord(xclic),absycoord(yclic),scale-SCALEMIN,10<<(scale-SCALEMIN));
			if (station>=0)
			{
				interfActive(ITF_STATION,1);
				istation->redrawbar(station);
			}
		}
*/	}
}


void Ctr_map::zoom(int change)
{
	if (zoomk) return;
	if (change) zoomdir=-zoomdir;
	if ((scale+zoomdir>SCALEMAX)||(scale+zoomdir<SCALEMIN))
	{
		zoomrun=0;
		return;
	}
	if ((scale+zoomdir==SCALEHIGH)||(scale==SCALEHIGH))
	{
		setscale(scale+zoomdir);
		xclic=-1;
		zoomrun=0;
		if (scale==SCALEHIGH) shell->setenable(0);
		else shell->setenable(1);
		return;
	}
	else if (zoomdir<0)
	{
		zoomk=1;
		xclic=-1;
		zoomrun=1;
	}
	else
	{
		setscale(scale+1);
		zoomk=255;
		xclic=-1;
		zoomrun=1;
	}
	shell->setenable(0);
	lines->setenable(0);
	gps->setenable(0);
}

void Ctr_map::zoomd(int in)
{
	if (in) zoomdir=-1;
	else zoomdir=1;
	zoom(0);
}

void Ctr_map::zoomstop()
{
	zoomrun=0;
}


int Ctr_map::getscale()
{
	return scale-SCALEMIN;
}

int Ctr_map::getgpsmode()
{
	if (!gpsstate) return 0;
	if (gpsstate<0) return 3;
	if (gpstrack) return 2;
	return 1;
}

void Ctr_map::getcoord(int* x,int *y)
{
	*x=absxcoord((*x)-xs);
	*y=absycoord((*y)-ys);
}


Ctr_map* Ctr_map_create(Container *c,Control* f,int x,int y,int w,int h,int wm,int hm)
{
	return new Ctr_map(c,f,x,y,w,h,wm,hm);
}

