//--------------------------
// TAO - 2006 - Sylvain Huet
//--------------------------

#include <windows.h>
#include <stdio.h>
#include <string.h>

#include "param.h"
#include "hash.h"
#include "file.h"


const uchar inv8[128]=
{
1,171,205,183,57,163,197,239,241,27,61,167,41,19,53,223,225,139,173,151,25,131,165,207,209,251,29,135,9,243,21,191,193,107,141,119,249,99,133,175,177,219,253,103,233,211,245,159,161,75,109,87,217,67,101,143,145,187,221,71,201,179,213,127,129,43,77,55,185,35,69,111,113,155,189,39,169,147,181,95,97,11,45,23,153,3,37,79,81,123,157,7,137,115,149,63,65,235,13,247,121,227,5,47,49,91,125,231,105,83,117,31,33,203,237,215,89,195,229,15,17,59,93,199,73,51,85,255
};

int encode8(uchar* src,int len,uchar key,uchar alpha)
{
	while(len--)
	{
		uchar v=*src;
		*(src++)=alpha+(v*inv8[key>>1]);
		key=v+v+1;
	}
	return key;
}


void strcpyspe(char* dst,char *src,int len)
{
	int i,j;
	j=0;
	for(i=0;i<len;i++)
	{
		if ((src[i]=='\\')&&(src[i+1]=='n'))
		{
			i++;
			dst[j++]=10;
		}
		else dst[j++]=src[i];
	}
	dst[j]=0;
}

HashKey::HashKey(uchar* k,int klen,uchar *v,int vlen,HashKey* nxt)
{
	next=nxt;
	key=new char[klen+1]; memcpy(key,(char*)k,klen);key[klen]=0;
	val=new char[vlen+1]; strcpyspe(val,(char*)v,vlen);
}

HashKey::~HashKey()
{
	if (next) delete next;
	delete key;
	delete val;
}

void HashKey::dump()
{
	PRINTF("]%s:%s\n",key,val);
	if (next) next->dump();
}



Hash::Hash(int nb)
{
	n=nb;
	htab=(HashKey**) new int[n];
	for(int i=0;i<n;i++) htab[i]=NULL;
}
Hash::~Hash()
{
	for(int i=0;i<n;i++) if (htab[i]) delete htab[i];
	if (htab) delete htab;
}

int	Hash::hashval(uchar *key,int len)
{
	int k=0;
	if (n<2) return k;
	while((len--)&&(*key)) k+=*(key++);
	return k%n;
}

int Hash::parse(uchar *src)
{
	int i,name_b,name_e,val_b,val_e,n;

	n=0;
	i=0;
	while(src[i])
	{
		while((src[i])&&(src[i]<=32)) i++;
		if (!src[i]) return n;
		name_b=i;
		while(src[i]>32) i++;
		if (!src[i]) return n;
		name_e=i;
		while((src[i])&&((src[i]==32)||(src[i]==9))) i++;
		if (!src[i]) return n;
		int hash=hashval(src+name_b,name_e-name_b);
		if (src[i]>32)	// il y a bien un argument
		{
			val_b=i;
			while(src[i]>=32) i++;
			val_e=i;
			htab[hash]=new HashKey(src+name_b,name_e-name_b,src+val_b,val_e-val_b,htab[hash]);
		}
		else
			htab[hash]=new HashKey(src+name_b,name_e-name_b,src,0,htab[hash]);
		n++;
	}
	return n;
}

HashKey* hashreverse(HashKey* pnt)
{
	HashKey* last=NULL;
	HashKey* next=NULL;
	while(pnt)
	{
		next=pnt->next;
		pnt->next=last;
		if (!next) return pnt;
		last=pnt;
		pnt=next;
	}
	return NULL;
}

void Hash::reverse()
{
	for(int i=0;i<n;i++) htab[i]=hashreverse(htab[i]);
}

int Hash::load(char *filename)
{
	return load(filename,0);
}

int Hash::load(char *filename,int crypt)
{
	int size;
	char* src=fileLoad(filename,&size);
	if (!src) return -1;
	if (crypt) encode8((uchar*)src,size,0x45,45);
	int n=parse((uchar *)src);
	reverse();
	fileRelease(src);
	return n;
}


void Hash::dump()
{
	for(int i=0;i<n;i++)
	{
		PRINTF("slot %d\n",i);
		if (htab[i]) htab[i]->dump();
	}
}

char* Hash::getnth(char *key,int n)
{
	if (!key) return NULL;
	int hash=hashval((uchar *)key,strlen(key));
	HashKey* p=htab[hash];
	while(p)
	{
		if ((!strcmp(key,p->key))&&(n--<=0)) return p->val;
		p=p->next;
	}
	return "";
}

char* Hash::get(char *key)
{
	return getnth(key,0);
}

HashKey* Hash::getnth(int n)
{
	HashKey* p=htab[0];
	while(n--) if (p) p=p->next;
	return p;
}
