//--------------------------
// TAO - 2006 - Sylvain Huet
//--------------------------

#include <windows.h>
#include <windowsx.h>
#include <wingdi.h>
#include <commctrl.h>
#include <commdlg.h>
#include <stdio.h>
#include <string.h>

#include "param.h"
#include "metro.h"
#include "file.h"


Metro::Metro()
{
	nsort=nquais=nlignes=nstations=0;
	data=NULL;
	sort=NULL;
	val=NULL;
	prec=NULL;
	sortin=NULL;
	used=NULL;
	indindex=NULL;
	dataindex=NULL;
	sel=NULL;
}

Metro::~Metro()
{
	if (data) fileRelease(data);
	if (sort) delete sort;
	if (val) delete val;
	if (prec) delete prec;
	if (sortin) delete sortin;
	if (used) delete used;
	if (sel) delete sel;

}

int Metro::msbtoint(char *p)
{
	return ((p[0]&255)<<24)+((p[1]&255)<<16)+((p[2]&255)<<8)+(p[3]&255);
}

int Metro::msbtoshort(char *p)
{
	return ((p[0]&255)<<8)+(p[1]&255);
}

int Metro::load(char *filename)
{
	data=fileLoad(filename);
	if (!data) return -1;

	int i=0;

	nlignes=msbtoint(data+i); i+=4;
	indlignes=(int*)(&data[i]); i+=(nlignes<<2);

	nstations=msbtoint(data+i); i+=4;
	indstations=(int*)(&data[i]); i+=(nstations<<2);
	xstations=(int*)(&data[i]); i+=(nstations<<2);
	ystations=(int*)(&data[i]); i+=(nstations<<2);

	nquais=msbtoint(data+i); i+=4;
	indquais=(int*)(&data[i]); i+=((nquais+1)<<2);
	dataquais=(char*)(&data[i]); i+=msbtoint((char*)&indquais[nquais]);
	int k=msbtoint(data+i); i+=4;
	datanames=(char*)(&data[i]); i+=k;
	k=msbtoint(data+i); i+=4;
	hotspots=(char*)(&data[i]); i+=k;

	indindex=(int*)(&data[i]); i+=(nstations<<2);
	int lendataindex=msbtoint(data+i); i+=4;
	dataindex=(char*)(&data[i]);i+=lendataindex;

	sort=new short[nquais+1];
	val=new short[nquais+1];
	prec=new short[nquais+1];
	sortin=new char[nquais+1];
	used=new char[nquais+1];
	sel=new char[nstations];


	PRINTF("load metro done\n");
/*
	for(i=0;i<nstations;i++)
	{
		PRINTF("%s %dx%d\n",getstationname(i),getstationx(i),getstationy(i));
	}
*/
//	solve(296,190);
//	solve(273,293);
//	solve(217,153);
//	solve(245,244);
//	solve(25,15);
	return 0;
}

int Metro::select(char* key)
{
	int i;
	int n=0;
	for(i=0;i<nstations;i++)
	{
		int k=0;
		if ((strstr(dataindex+msbtoint((char*)&indindex[i]),key)!=NULL)) k=1;
		sel[i]=k;
		n+=k;
//		if (k) PRINTF("-%s : %s\n",&index[tpoi[i].index],getdesc(&tpoi[i]));
	}
	return n;

}

int Metro::getnstations()
{
	return nstations;
}


int Metro::hotspot(int x,int y,int scale,int dist)
{
	int s;
	int i=0;
	while((s=msbtoint(hotspots+i))>=0)
	{
		i+=4;
		int n=msbtoint(hotspots+i);i+=4;
		if (s==scale)
		{
			int j=0;
			while(j<n)
			{
				int xh=msbtoint(hotspots+i+j);j+=4;
				int yh=msbtoint(hotspots+i+j);j+=4;
				int xx=x-xh;
				int yy=y-yh;
				if ((xx>=-dist)&&(xx<=dist)&&(yy>=-dist)&&(yy<=dist))
				{
					PRINTF("OK ! %dx%d\n",xh,yh);
					int mind;
					int candidate=-1;
					for(int k=0;k<nstations;k++)
					{
						xx=getstationx(k)-xh;
						yy=getstationy(k)-yh;
						if (xx<0)xx=-xx;
						if (yy<0)yy=-yy;
						if (xx<yy) xx=yy;
						if ((candidate<0)||(xx<mind))
						{
							candidate=k;
							mind=xx;
						}
					}
					if (candidate>=0) PRINTF("%s\n",getstationname(candidate));
					return candidate;
				}
			}
		}
		i+=n;
	}
	PRINTF("No metro\n");
	return -1;
}


char* Metro::getlignename(int i)
{
	if ((i<0)||(i>=nlignes)) return NULL;
	return datanames+msbtoint((char*)&indlignes[i]);
}

char* lignesname[NBLIGNES]=
{
"1","2","3","3bis","4","5","6","7","7bis","8","9","10","11","12","13","14","A","B","C"
};
int lignescolor[NBLIGNES*3]=
{
	255,191,0,11,74,158,138,120,15,110,191,185,184,38,143,
	253,121,41,110,186,111,249,137,173,110,186,111,
	184,139,194,170,185,14,210,160,25,117,61,10,
	1,106,60,110,191,185,75,0,125,254,55,15,
	83,125,185,232,206,16
};

int Metro::isrer(int i)
{
	char* p=getlignename(i);
	if (!p) return 0;
	char c=p[0];
	if ((c>='A')&&(c<='E')) return 1;
	return 0;
}
int Metro::getlinecolor(int i)
{
	i=getindligne(i);
	if (i<0) return 0xffff;
	i*=3;
	return RGBTO16(lignescolor[i],lignescolor[i+1],lignescolor[i+2]);
}

char* Metro::getstationname(int i)
{
	if ((i<0)||(i>=nstations)) return NULL;
	return datanames+msbtoint((char*)&indstations[i]);
}

int Metro::isselected(int i)
{
	if ((i<0)||(i>=nstations)) return 0;
	return sel[i];
}

int Metro::getstationx(int i)
{
	if ((i<0)||(i>=nstations)) return 0;
	return msbtoint((char*)&xstations[i]);
}

int Metro::getstationy(int i)
{
	if ((i<0)||(i>=nstations)) return 0;
	return msbtoint((char*)&ystations[i]);
}

int Metro::remsort()
{
	if (nsort<=0) return -1;
	int r=sort[0];
	sortin[r]=0;
	nsort--;
	for(int i=0;i<nsort;i++) sort[i]=sort[i+1];
	return r;
}

void Metro::remsort(int v)
{
	int i;
	for(i=0;i<nsort;i++) if (sort[i]==v)
	{
		int j;
		for(j=i+1;j<nsort;j++) sort[j-1]=sort[j];
		nsort--;
		sortin[v]=0;
		return;
	}
}

void Metro::insertsort(int q)
{
	int i=nsort-1;
	if (sortin[q]) {PRINTF("erreur insertsort !!\n"); return ; }
	while((i>=0)&&(val[sort[i]]>val[q]))
	{
		sort[i+1]=sort[i];
		i--;
	}
	sort[i+1]=q;
	sortin[q]=1;
	nsort++;
}

void Metro::dumpsort()
{
	int i,j;
	i=0;
	if (nsort>10) i=nsort-10;
	for(;i<nsort;i++)
	{
		PRINTF("%d.%d = %d\n",i,sort[i],val[sort[i]]);
	}
}

int Metro::stationfromquai(int i)
{
	return msbtoshort(dataquais+msbtoint((char*)&indquais[i])+2);
}

int Metro::lignefromquai(int i)
{
	return msbtoshort(dataquais+msbtoint((char*)&indquais[i]));
}

int Metro::corrfromquai(int q,int n)
{
	return msbtoshort(dataquais+msbtoint((char*)&indquais[q])+4+2*n);
}

void Metro::tryq(int u,int v,int k)
{
	if ( ((!used[v])&&(!sortin[v]))|| (val[v]>val[u]+k))
	{
		val[v]=val[u]+k;
		prec[v]=u;
		if (sortin[v]) remsort(v);
		insertsort(v);
	}
}

/*
int Metro::solve(int a,int b)
{
	nsort=0;
	if ((a<0)||(a>=nstations)) return 0;
	if ((b<0)||(b>=nstations)) return 0;

	int i;
	for(i=0;i<nquais;i++) sortin[i]=used[i]=0;
	for(i=0;i<nquais;i++) val[i]=0;
	for(i=0;i<nquais;i++) prec[i]=-1;
	nsort=0;
	for(i=0;i<nquais;i++) if (stationfromquai(i)==a) insertsort(i);

	int u;
	while((u=remsort())!=-1)
	{
		used[u]=1;
		if (stationfromquai(u)==b)
		{
			PRINTF("found !! %d min\n",val[u]);
			int j=nquais;
			while(u!=-1)
			{
				sort[--j]=u;
				u=prec[u];
			}
			nsort=0;
			while(j<nquais) sort[nsort++]=sort[j++];
			return nsort;
		}
		int v=u+1;
		if (lignefromquai(v)==lignefromquai(u)) tryq(u,v,1);

		i=0;
		while((v=corrfromquai(u,i++))!=65535) tryq(u,v,5);

	}
	nsort=0;
	PRINTF("not found\n");
	return 0;
}
*/
int Metro::dist(int xo,int yo,int xd,int yd)
{
	int dx=xd-xo;
	int dy=yd-yo;
	if (dx<0) dx=-dx;
	if (dy<0) dy=-dy;
	return dx+dy;
}
int Metro::dist(int q,int x,int y)
{
	int i=stationfromquai(q);
	return dist(getstationx(i),getstationy(i),x,y);
}
#define DIV_MARCHE 50
#define T_METRO 1
#define T_RER 3
#define T_CHGRAPIDE 5
#define T_CHGLENT 10


int Metro::solve(int xo,int yo,int xd,int yd)
{
	nsort=0;
	timetodest=-1;
	int i;
	for(i=0;i<=nquais;i++) sortin[i]=used[i]=0;
	for(i=0;i<=nquais;i++) val[i]=0;
	for(i=0;i<=nquais;i++) prec[i]=-1;
	nsort=0;
	for(i=0;i<nquais;i++)
	{
		int d=dist(i,xo,yo);
		if (d<5000)
		{
			PRINTF(">depart %s %d\n",getstationname(stationfromquai(i)),d);
			val[i]=d/DIV_MARCHE;
			insertsort(i);
		}
	}

	int u;
	while((u=remsort())!=-1)
	{
		used[u]=1;
		if (u==nquais)
		{
			PRINTF("found !! %d min\n",val[u]);
			int d=dist(xo,yo,xd,yd)/DIV_MARCHE;
			if (d<=val[u]+1)
			{
				PRINTF("tout faire  pied %d mn\n",d);
				sort[0]=u;
				nsort=1;
				val[u]=d;
				timetodest=d;
				return nsort;
			}
			int j=nquais;
			while(u!=-1)
			{
				sort[--j]=u;
				u=prec[u];
			}
			nsort=0;
			while(j<nquais) sort[nsort++]=sort[j++];
			for(i=0;i<nsort;i++)
				if (sort[i]==nquais) PRINTF("finir  pied\n");
				else PRINTF(">%s     (%s) %d mn\n",getstationname(stationfromquai(sort[i])),getlignename(lignefromquai(sort[i])),val[sort[i]]);
			timetodest=val[nquais];
			return nsort;
		}
		int v=u+1;
		int l=lignefromquai(u);
		if (lignefromquai(v)==l)
		{
			tryq(u,v,isrer(l)?T_RER:T_METRO);
		}
		int d=dist(u,xd,yd);
		if (d<5000)
		{
			tryq(u,nquais,d/DIV_MARCHE);
		}

		i=0;
		while((v=corrfromquai(u,i++))!=65535) tryq(u,v,(isrer(l)==isrer(lignefromquai(v)))?T_CHGRAPIDE:T_CHGLENT);

	}
	nsort=0;
	PRINTF("not found\n");
	return 0;
}

int Metro::getnthquai(int i)
{
	if ((i<0)||(i>=nsort)) return -1;
	return sort[i];
}

int Metro::getnsteps()
{
	return nsort;
}

int Metro::gettime()
{
	return timetodest;
}

int Metro::getindligne(int i)
{
	char* p=getlignename(i);
	char buff[8];

	if (!p) return -1;
	i=0;
	while((i<6)&&(p[i]!=32)) buff[i++]=p[i];
	buff[i]=0;
	for(i=0;i<NBLIGNES;i++) if (!strcmp(buff,lignesname[i])) return i;
	return -1;
}

int Metro::lignesfromstation(int i)
{
	int k;
	for(k=0;k<NBLIGNES;k++) stationslignes[k]=0;
	if ((i<0)||(i>=nstations)) return 0;
	for(k=0;k<nquais;k++)
		if (stationfromquai(k)==i)
		{
			int l=getindligne(lignefromquai(k));
//			PRINTF("found quai %d, ligne %d\n",k,l);
			if (l>=0) stationslignes[l]=1;
		}
	return 0;
}

