//--------------------------
// TAO - 2006 - Sylvain Huet
//--------------------------

#include <windows.h>
#include <windowsx.h>
#include <wingdi.h>
#include <commctrl.h>
#include <commdlg.h>
#include <string.h>
#include<stdio.h>

#include "param.h"
#include "config.h"
#include "poi.h"
#include "file.h"
#include "control.h"
#include "interf.h"

#define POI_RESET_TIMEOUT (1000*60*3)

Poilist::Poilist()
{
	npoi=0;
	tpoi=NULL;
	ncrit=0;
	tcrit=NULL;
	names=NULL;
	desc=NULL;
	index=NULL;
	curtype=0;
	notnul=0;
	poistart=NULL;
}

Poilist::~Poilist()
{
	if (tpoi) delete tpoi;
	if (tcrit) delete tcrit;
	if (names) delete names;
	if (desc) delete desc;
	if (index) delete index;
}


int Poilist::msbtoint(char *p)
{
	return ((p[0]&255)<<24)+((p[1]&255)<<16)+((p[2]&255)<<8)+(p[3]&255);
}

int Poilist::load(char *filename)
{
	char* data=fileLoad(filename);
	if (!data) return -1;

	poistart=NULL;

	int i=0;
	npoi=msbtoint(data+i); i+=4;

	if (tpoi) delete tpoi;
	tpoi=new Poi[npoi];

	int j;
	for(j=0;j<npoi;j++) tpoi[j].id=msbtoint(data+i+(j<<2));
	i+=npoi*4;
	for(j=0;j<npoi;j++) tpoi[j].desc=msbtoint(data+i+(j<<2));
	i+=npoi*4;
	for(j=0;j<npoi;j++) tpoi[j].index=msbtoint(data+i+(j<<2));
	i+=npoi*4;
	for(j=0;j<npoi;j++) tpoi[j].tp=msbtoint(data+i+(j<<2));
	i+=npoi*4;
	for(j=0;j<npoi;j++) tpoi[j].x=msbtoint(data+i+(j<<2));
	i+=npoi*4;
	for(j=0;j<npoi;j++) tpoi[j].y=msbtoint(data+i+(j<<2));
	i+=npoi*4;

	for(j=0;j<npoi;j++) tpoi[j].zoom=msbtoint(data+i+(j<<2));
	i+=npoi*4;
	for(j=0;j<npoi;j++) tpoi[j].meta=msbtoint(data+i+(j<<2));
	i+=npoi*4;
	for(j=0;j<npoi;j++) tpoi[j].cat=msbtoint(data+i+(j<<2));
	i+=npoi*4;
	for(j=0;j<npoi;j++) tpoi[j].icon=msbtoint(data+i+(j<<2));
	i+=npoi*4;
	for(j=0;j<npoi;j++) tpoi[j].audioonly=msbtoint(data+i+(j<<2));
	i+=npoi*4;
	for(j=0;j<npoi;j++) tpoi[j].perso=msbtoint(data+i+(j<<2));
	i+=npoi*4;


	ncrit=msbtoint(data+i); i+=4;
	tcrit=new char[ncrit*npoi];
	memcpy(tcrit,data+i,ncrit*npoi); i+=ncrit*npoi;

	int lennames=msbtoint(data+i); i+=4;
	names=new char[lennames];
	memcpy(names,data+i,lennames);
	i+=lennames;

	int lendesc=msbtoint(data+i); i+=4;
	PRINTF("lendesc=%d\n",lendesc);
	desc=new char[lendesc];
	memcpy(desc,data+i,lendesc);
	i+=lendesc;

	int lenindex=msbtoint(data+i); i+=4;
	PRINTF("lenindex=%d\n",lenindex);
	index=new char[lenindex];
	memcpy(index,data+i,lenindex);
	i+=lenindex;

	fileRelease(data);

	char *start=config->get("start");
//	PRINTF("poistart= %s\n",start);
	for(i=0;i<npoi;i++)
	{
		tpoi[i].num=i;
		if (!strcmp(start,getname(&tpoi[i]))) poistart=&tpoi[i];
	}

//	for(i=0;i<npoi;i++) PRINTF("%d %s %d\n",i,getname(&tpoi[i]),tpoi[i].meta);

	return 0;
}
void Poilist::touch()
{
	lasttimechg=GetTickCount();
}

int Poilist::nbsorted()
{
	return notnul;
}
Poi* Poilist::nthsorted(int i)
{
	if ((i<0)||(i>=notnul)) return NULL;
	return &tpoi[tpoi[i].sort];
}

int Poilist::getbest(Poi** tb,int nmax,int xmin,int xmax,int ymin,int ymax,int scale,int dist)
{
	int i,j;
	j=0;
	i=0;
//	PRINTF("get best %d,%d - %d,%d\n",xmin,xmax,ymin,ymax);
	while((j<nmax)&&(i<notnul))
	{
//		PRINTF("%d -",i);
		Poi* p=&tpoi[tpoi[i].sort];
//		PRINTF("chk poi %d -> %d\n",tpoi[i].sort,p->score);
		if ((curtype==tpoi[i].sort+1000)||
			((p->score)&&(p->zoom>=scale)&&(p->x ||p->y)&&(p->x>=xmin)&&(p->x<xmax)&&(p->y>=ymin)&&(p->y<ymax)) )
		{
			int h=0;
			for(int k=0;k<j;k++)
			{
				Poi* q=tb[k];
				int d=(((q->icon>=1000)?16:12)+((p->icon>=1000)?16:12))<<dist;
				if ((p->x>q->x-d)&&(p->x<q->x+d)&&(p->y>q->y-d)&&(p->y<q->y+d)) h=1;
			}
			if (!h) tb[j++]=p;
		}
		i++;
	}
	return j;
}


void Poilist::clearscore()
{
	for(int i=0;i<npoi;i++) tpoi[i].score=0;
	touch();
}

void Poilist::mergecritt(int n,int coeff)
{
	char* p=tcrit+n;
	for(int i=0;i<npoi;i++)
	{
		int s=(*p)&255;
//		if (s) s+=10;
//		PRINTF("%s:%d\n",getname(&tpoi[i]),s);
		if (s>tpoi[i].score)tpoi[i].score=s;
		p+=ncrit;
	}
	touch();
}

void Poilist::mergetype(int tp,int val)
{
	for(int i=0;i<npoi;i++) if (tpoi[i].tp==tp)
	{
		if (val>0) tpoi[i].score=val;
		else tpoi[i].score=tcrit[i*ncrit+ncrit-1];
//		PRINTF("%s:%d\n",getname(&tpoi[i]),tpoi[i].score);
	}
	touch();
}

void Poilist::currentcrit(int n,int addstart)
{
	clearscore();
/*
			mergetype(0,1);
			mergetype(1,1);
			mergetype(2,1);
*/
	curtype=n;
	if (n>=0)
	{
		switch(n)
		{
		case 0:
			mergecritt(ncrit-1,1);
			break;
		case 1:
			break;
		case 2:
			mergecritt(132,1);
			mergecritt(44,1);
			mergetype(0,0);
			break;
		case 3:
			mergecritt(21,1);
			mergecritt(22,1);
			mergecritt(23,1);
			break;
		case 4:
			mergecritt(0,1);
			mergecritt(1,1);
			mergecritt(2,1);
			mergecritt(4,1);
			mergecritt(5,1);
			break;
		}
	}
	else
	{
		mergecritt(-n-1,1);
	}

	int i;
	for(i=0;i<npoi;i++) if (tpoi[i].perso) tpoi[i].score=0;
	if ((addstart)&&(poistart)) poistart->score=101;
	sort();
}

int Poilist::getpoix(Poi* p)
{
	if ((p->x)||(p->y)) return p->x;
	if (p->meta>=0)
	{
		Poi* q=poilist->getnth(p->meta);
		return q->x;
	}
	return 0;
}
int Poilist::getpoiy(Poi* p)
{
	if ((p->x)||(p->y)) return p->y;
	if (p->meta>=0)
	{
		Poi* q=poilist->getnth(p->meta);
		return q->y;
	}
	return 0;
}

int Poilist::conc(int p,int q)
{
	if (p==-1) return q;
	int i=p;
	while(tpoi[i].tl!=-1) i=tpoi[i].tl;
	tpoi[i].tl=q;
	return p;
}

void Poilist::divide(int p,int *q,int *r)
{
	*q=*r=-1;
	if (p==-1) return;
	int x=tpoi[p].score;
	while(p!=-1)
	{
		int np=tpoi[p].tl;
		if (tpoi[p].score>=x)
		{
			tpoi[p].tl=*q;
			*q=p;
		}
		else
		{
			tpoi[p].tl=*r;
			*r=p;
		}
		p=np;
	}
}

void Poilist::dump(int p)
{
	while(p!=-1)
	{
		PRINTF("%d[%d]::",p,tpoi[p].score);
		p=tpoi[p].tl;
	}
	PRINTF("nil\n");
}

int Poilist::quicksort(int p)
{
	int q,r;
//	PRINTF("quicksort :");dump(p);
	if (p==-1) return -1;
	if (tpoi[p].tl==-1) return p;
	divide(p,&q,&r);
	if (q==-1) return r;
	if (r==-1) return q;
	q=quicksort(q);
	r=quicksort(r);
	return conc(q,r);
}


void Poilist::sort()
{
	int i;
	notnul=0;
	int nul=npoi;

	touch();

	int hd=-1;
	int hdnul=-1;

//	for(i=0;i<npoi;i++) tpoi[i].score=(i<20)?tpoi[i].score:0;

	for(i=0;i<npoi;i++)
	{
		if (tpoi[i].score)
		{
			if (tpoi[i].icon>=1000) tpoi[i].score+=1000;
			tpoi[i].tl=hd;
			hd=i;
		}
		else
		{
			tpoi[i].tl=hdnul;
			hdnul=i;
		}
	}

	//quicksort sur hd
	hd=quicksort(hd);
//	PRINTF("result :");dump(hd);

	notnul=0;
	while(hd!=-1)
	{
		tpoi[notnul++].sort=hd;
		hd=tpoi[hd].tl;
	}
	int nuls=npoi;
	while(hdnul!=-1)
	{
		tpoi[--nuls].sort=hdnul;
		hdnul=tpoi[hdnul].tl;
	}


	/*
	for(i=0;i<npoi;i++)
	{
		if (tpoi[i].score) tpoi[notnul++].sort=i;
		else tpoi[--nul].sort=i;
	}
	int cont=1;
	while(cont)
	{
		cont=0;
		for(i=0;i<notnul-1;i++)
			if (
				( (tpoi[tpoi[i].sort].score<tpoi[tpoi[i+1].sort].score)&&((tpoi[tpoi[i].sort].icon<1000)||(tpoi[tpoi[i+1].sort].icon>=1000)))
				||((tpoi[tpoi[i].sort].icon<1000)&&(tpoi[tpoi[i+1].sort].icon>=1000)))
			{
				int k=tpoi[i].sort;
				tpoi[i].sort=tpoi[i+1].sort;
				tpoi[i+1].sort=k;
				cont=1;
			}
	}
*/
}



void Poilist::setcrit(int n,int addstart)
{
	currentcrit(-1-n,addstart);
}

void Poilist::setsimple(int n,int addstart)
{
	currentcrit(n,addstart);
}

void Poilist::setpoi(int poi)
{
	poilist->currentcrit(1,1);
	poilist->setscore(&tpoi[poi],102);
	poilist->sort();
	curtype=1000+poi;

}

int Poilist::getcurrenttype()
{
	return curtype;
}

char* Poilist::getname(Poi* p)
{
	return &names[p->id];
}


char* Poilist::getdesclocal(Poi* p)
{
	return &desc[p->desc];
}

char* Poilist::getdesclocalized(Poi* p)
{
	return getdescn(p,1000);
}

char* Poilist::getdescn(Poi* p,int n)
{
	char* d=&desc[p->desc];
	if ((*d)==0) return d;
	char *lastd=d;
	while(n>1)
	{
		d+=strlen(d)+1;
		if ((*d)==0) return lastd;
		lastd=d;
		n--;
	}
	return lastd;
}

int Poilist::getnpoi()
{
	return npoi;
}
Poi* Poilist::getnth(int i)
{
	if ((i<0)||(i>=npoi)) return NULL;
	return tpoi+i;
}

Poi* Poilist::getstart()
{
	return poistart;
}

void Poilist::setscore(Poi* p,int i)
{
	p->score=i;
}
int Poilist::getscore(Poi* p)
{
	return p->score;
}

int Poilist::score(int poi,int i)
{
	if ((poi<0)||(poi>=npoi))
	{
		PRINTF("mauvais poi\n");
		return 0;
	}
	if ((i<0)||(i>=ncrit))
	{
		PRINTF("mauvais critere\n");
		return 0;
	}
	return tcrit[poi*ncrit+i];
}

int multiplestr(char* content,char* key)
{
	int n=1;
	while(*content)
	{
		if (strstr(content,key)!=NULL) return n;
		content+=strlen(content)+1;
		n++;
	}
	return 0;
}

void dumpindex(char* content)
{
	while(*content)
	{
		printf("{%s}",content);
		content+=strlen(content)+1;
	}
}
int Poilist::select(char* key)
{
	int i;
	int n=0;
	for(i=0;i<npoi;i++)
	{
		int k=0;
		if ((!tpoi[i].perso)||(&tpoi[i]==poistart))
		{
			if (!tpoi[i].audioonly) k=multiplestr(&index[tpoi[i].index],key);
		}
//		if ((!tpoi[i].audioonly)&&(strstr(&index[tpoi[i].index],key)!=NULL)) k=1;
		tpoi[i].select=k;
		if (k) n++;
/*		if (k)
		{
			dumpindex(&index[tpoi[i].index]);
			PRINTF("found %s -in- %s : %s [%s]\n",key,&index[tpoi[i].index],getdesclocal(&tpoi[i]),getdesclocalized(&tpoi[i]));
		}
*/	}
	return n;

}

void Poilist::checkreset()
{
/*	if (GetTickCount()-lasttimechg>POI_RESET_TIMEOUT)
	{
		touch();
		if (getcurrenttype()) setsimple(0,1);
		imap->updatepoi();
		isimple->redrawbar();
//		imap->resetpoi();
	}
*/
}

Poilist* poilist=NULL;

int poilistInit()
{
	poilist=new Poilist();
	poilist->load(ilang->loc("loc_poilist_data"));
	poilist->currentcrit(0,1);
	return 0;
}


int poilistReload()
{

	ipanier->listclear();
	if (poilist) delete poilist;
	poilist=new Poilist();
	poilist->load(ilang->loc("loc_poilist_data"));
	poilist->setsimple(0,1);
	imap->resetpoi();
	isimple->redrawbar();
	interfActive(ITF_SIMPLE,0);
	return 0;
}